"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const tslib_1 = require("tslib");
const ipfsAPI = require('ipfs-api');
const ipld = require('ipld-dag-cbor');
var jiff = require('jiff');
const _ = require("lodash");
class TransmuteIpfsSingleton {
    constructor() {
        this.isObjectStore = true;
        this.hashesToPatches = (hashes) => tslib_1.__awaiter(this, void 0, void 0, function* () {
            let marshalledPatches = yield Promise.all(hashes.map((hash) => {
                return this.readBuffer(hash);
            }));
            let promises = yield marshalledPatches.map((marshalled) => {
                return new Promise((resolve, reject) => {
                    ipld.util.deserialize(new Buffer(marshalled), (err, data) => {
                        if (err) {
                            reject(err);
                        }
                        resolve(data);
                    });
                });
            });
            return Promise.all(promises);
        });
        this.applyPatches = (obj, patches) => {
            let patched = _.cloneDeep(obj);
            patches.forEach((patch) => {
                patched = jiff.patchInPlace(patch, patched);
            });
            return patched;
        };
        this.applyIPLDHashes = (obj, hashes) => tslib_1.__awaiter(this, void 0, void 0, function* () {
            let patches = yield this.hashesToPatches(hashes);
            return this.applyPatches(obj, patches);
        });
        this.patch = (target, patch) => {
            return jiff.patchInPlace(patch, target);
        };
        this.diff = (start, end) => {
            return jiff.diff(start, end);
        };
    }
    init(config = {
            host: 'localhost',
            port: '5001',
            options: {
                protocol: 'http'
            }
        }) {
        this.config = config;
        this.ipfs = ipfsAPI(this.config.host, this.config.port, this.config.options);
        return this;
    }
    addFromFs(folderPath, options = {
            recursive: true,
            ignore: ['node_modules/**']
        }) {
        return new Promise((resolve, reject) => {
            this.ipfs.util.addFromFs(folderPath, options, (err, result) => {
                if (err) {
                    reject(err);
                }
                resolve(result);
            });
        });
    }
    addFromURL(url) {
        return new Promise((resolve, reject) => {
            this.ipfs.util.addFromURL(url, (err, result) => {
                if (err) {
                    reject(err);
                }
                resolve(result);
            });
        });
    }
    writeBuffer(buffer) {
        return new Promise((resolve, reject) => {
            this.ipfs.add(buffer, (err, res) => {
                if (err || !res) {
                    reject(err);
                }
                resolve(res[0].path);
            });
        });
    }
    writeObject(obj) {
        return new Promise((resolve, reject) => {
            ipld.util.serialize(obj, (err, serialized) => {
                if (err) {
                    reject(err);
                }
                resolve(this.writeBuffer(serialized));
            });
        });
    }
    readBuffer(hash) {
        return new Promise((resolve, reject) => {
            this.ipfs.cat(hash, (err, stream) => {
                if (err) {
                    reject(err);
                }
                let data = new Buffer('');
                stream.on('data', (chunk) => {
                    data = Buffer.concat([data, chunk]);
                });
                stream.on('end', () => {
                    resolve(data);
                });
            });
        });
    }
    readObject(path) {
        if (path.indexOf('ipfs/') !== -1) {
            path = path.split('ipfs/')[1];
        }
        return this.readBuffer(path)
            .then((serialized) => {
            return new Promise((resolve, reject) => {
                ipld.util.deserialize(serialized, (err, data) => {
                    if (err) {
                        reject(err);
                    }
                    resolve(data);
                });
            });
        });
    }
    statesToPatches(states) {
        return new Promise((resolve, reject) => {
            let patches = [];
            for (var i = 0; i <= states.length - 2; i++) {
                patches.push(jiff.diff(states[i], states[i + 1]));
            }
            resolve(patches);
        });
    }
    patchesToHashes(patches) {
        return new Promise((resolve, reject) => tslib_1.__awaiter(this, void 0, void 0, function* () {
            let marshalledPatches = patches.map((patch) => {
                return new Promise((resolve, reject) => {
                    ipld.util.serialize(patch, (err, serialized) => {
                        if (err) {
                            reject(err);
                        }
                        resolve(serialized);
                    });
                });
            });
            let buffers = yield Promise.all(marshalledPatches);
            let promises = buffers.map((mp) => {
                return this.writeBuffer(mp);
            });
            resolve(Promise.all(promises));
        }));
    }
}
exports.TransmuteIpfsSingleton = TransmuteIpfsSingleton;
exports.TransmuteIpfs = new TransmuteIpfsSingleton();
